from graph import BasicGraph


def subgraph(graph, nodes):
    """ Creates a subgraph as a copy from the graph
    :param graph: class Graph
    :param nodes: list of nodes
    :return: new instance of Graph.
    """
    assert isinstance(graph, BasicGraph)
    assert isinstance(nodes, list)
    G = object.__new__(graph.__class__)
    G.__init__()
    for n1 in nodes:
        G.add_node(n1)
        for n2 in graph.nodes(from_node=n1):
            G.add_edge(n1, n2, graph.edge(n1, n2))
    return G


def is_subgraph(graph1, graph2):
    """
    Checks is graph1 is subgraph in graph2
    :param graph1: instance of Graph
    :param graph2: instance of Graph
    :return: boolean
    """
    assert isinstance(graph1, BasicGraph)
    assert isinstance(graph2, BasicGraph)
    if not set(graph1.nodes()).issubset(set(graph2.nodes())):
        return False
    if not set(graph1.edges()).issubset(set(graph2.edges())):
        return False
    return True


def is_partite(graph, n):
    """ Checks if graph is n-partite
    :param graph: class Graph
    :param n: int, number of partitions.
    :return: boolean and partitions as dict[colour] = set(nodes) or None.
    """
    assert isinstance(graph, BasicGraph)
    assert isinstance(n, int)
    colours_and_nodes = {i: set() for i in range(n)}
    nodes_and_colours = {}
    n1 = set(graph.nodes()).pop()
    q = [n1]
    visited = set()
    colour = 0
    while q:
        n1 = q.pop()
        visited.add(n1)

        if n1 in nodes_and_colours:
            colour = nodes_and_colours[n1]
        else:
            colours_and_nodes[colour].add(n1)
            nodes_and_colours[n1] = colour

        next_colour = (colour + 1) % n
        neighbours = graph.nodes(from_node=n1) + graph.nodes(to_node=n1)
        for n2 in neighbours:
            if n2 in nodes_and_colours:
                if nodes_and_colours[n2] == colour:
                    return False, None
                # else:  pass  # it already has a colour and there is no conflict.
            else:  # if n2 not in nodes_and_colours:
                colours_and_nodes[next_colour].add(n2)
                nodes_and_colours[n2] = next_colour
                continue
            if n2 not in visited:
                q.append(n2)

    return True, colours_and_nodes


def same(path1, path2):
    """ Compares two paths to verify whether they're the same.
    :param path1: list of nodes.
    :param path2: list of nodes.
    :return: boolean.
    """
    start1 = path2.index(path1[0])
    checks = [
        path1[:len(path1) - start1] == path2[start1:],
        path1[len(path1) - start1:] == path2[:start1]
    ]
    if all(checks):
        return True
    return False


def has_path(graph, path):
    """ checks if path exists is graph
    :param graph: instance of Graph
    :param path: list of nodes
    :return: boolean
    """
    assert isinstance(graph, BasicGraph)
    assert isinstance(path, list)
    v1 = path[0]
    for v2 in path[1:]:
        if graph.edge(v1, v2) is None:
            return False
        else:
            v1 = v2
    return True


def has_cycles(graph):
    """ Checks whether the graph has a cycle
    :param graph: instance of class Graph.
    :return: bool
    """
    for n1, n2, d in graph.edges():
        if n1 == n2:
            return True
        if graph.depth_first_search(start=n2, end=n1):
            return True
    return False


def components(graph):
    """ Determines the components of the graph
    :param graph: instance of class Graph
    :return: list of sets of nodes. Each set is a component.
    """
    assert isinstance(graph, BasicGraph)
    nodes = set(graph.nodes())
    sets_of_components = []
    while nodes:
        new_component = set()
        sets_of_components.append(new_component)
        n = nodes.pop()  # select random node
        new_component.add(n)  # add it to the new component.

        new_nodes = set(graph.nodes(from_node=n))
        new_nodes.update(set(graph.nodes(to_node=n)))
        while new_nodes:
            n = new_nodes.pop()
            new_component.add(n)
            new_nodes.update(set(n for n in graph.nodes(from_node=n) if n not in new_component))
            new_nodes.update(set(n for n in graph.nodes(to_node=n) if n not in new_component))
        nodes = nodes - new_component
    return sets_of_components


def social_network_size(graph, n1, degrees_of_separation=None):
    """ Determines the nodes within a degree of separation range
    :param graph: Graph
    :param n1: start node
    :param degrees_of_separation: integer
    :return: set of nodes within given range
    """
    assert isinstance(graph, BasicGraph)
    assert n1 in graph.nodes()
    assert isinstance(degrees_of_separation, int)
    social_network = set(graph.nodes(from_node=n1))
    peer1 = social_network.copy()
    peer2 = set()
    scan_depth = 1
    while scan_depth < degrees_of_separation:
        scan_depth += 1
        for peer in peer1:
            new_peers = set(graph.nodes(from_node=peer)) - social_network
            peer2.update(new_peers)
            social_network.update(new_peers)
        peer1 = peer2
    return social_network


def phase_lines(graph):
    """ Determines the phase lines of a graph.
    :param graph: Graph
    :return: dictionary with node id : phase in cut.
    """
    if has_cycles(graph):
        raise ValueError("a cyclic graph will not have phaselines.")

    phases = {n: 0 for n in graph.nodes()}
    q = graph.nodes(in_degree=0)
    q_set = set(q)
    while q:
        n = q.pop(0)
        q_set.remove(n)
        level = phases[n]
        children = graph.nodes(from_node=n)
        for c in children:
            if phases[c] <= level:
                phases[c] = level + 1
            if c not in q_set:
                q.append(c)
                q_set.add(c)
    return phases


def sources(graph, n):
    """ Determines the set of sources of 'node' in a DAG
    :param graph: Graph
    :return: set of nodes
    """
    nodes = {n}
    q = [n]
    while q:
        new = q.pop(0)
        for src in graph.nodes(to_node=new):
            if src not in nodes:
                nodes.add(src)
            if src not in q:
                q.append(src)
    nodes.remove(n)
    return nodes







from itertools import count
from matplotlib import pyplot as plt
from graph import Graph
from graph.hash import graph_hash


# __all__ = ['Machine', 'Job', 'minimise_makespan', 'gantt']


class Process(object):
    __slots__ = ['id', 'run', 'setup', 'teardown']

    def __init__(self, id, run, setup, teardown):
        """
        :param run: int or float: runtime of the process.
        :param setup: int or float: setup time of the process.
        :param teardown: int or float: teardown time of the process.

        Note that consecutive executions of an already set-up process does
        not require to be set-up again hereby if: p = Process(2,3,5), then
        the makespan of  2 * p : 2 + 3 + 3 + 5 = 13 units of time.

        """
        self.id = id
        assert isinstance(run, (int, float)) and run >= 0
        self.run = run
        assert isinstance(setup, (int, float)) and setup >= 0
        self.setup = setup
        assert isinstance(teardown, (int, float)) and teardown >= 0
        self.teardown = teardown

    def copy(self):
        return Process(self.id, self.run, self.setup, self.teardown)


class Task(object):
    def __init__(self, process, job):
        self.process = process
        self.job = job

    def __lt__(self, other):
        return self.process.id < other.process.id

    def __repr__(self):
        return "{},{}".format(self.process, self.job)


class AppointmentError(Exception):
    pass


class Appointment(object):
    __slots__ = ['machine', 'job', 'process', 'earliest_start', 'start', 'finish']

    def __init__(self, machine, job, process, earliest_start, start=float('inf'), finish=float('inf')):
        """ An appointment between a machine and a job.
        :param machine: machine id
        :param job: job id
        :param process: process id
        :param earlist_start: estimated start
        :param start: start
        :param finish: finish
        """
        self.machine = machine
        self.job = job
        self.process = process
        self.earliest_start = earliest_start
        self.start = start
        self.finish = finish

    def copy(self):
        return Appointment(self.machine, self.job, self.process, self.earliest_start, self.start, self.finish)


class Machine(object):
    ids = count()

    def __init__(self):
        self.id = next(Machine.ids)
        self.processes = {}
        self.jobs = []
        self.schedule = []
        self.appointments = Graph()

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "{}-{}".format(self.__class__.__name__, self.id)

    def __hash__(self):
        return self.id

    def add_process(self, process_id, run_time, setup_time=0, teardown_time=0):
        self.processes[process_id] = Process(id=process_id, run=run_time, setup=setup_time, teardown=teardown_time)

    def makespan_finish(self):
        if not self.schedule:
            return 0
        return max([s.f for s in self.schedule])

    def evaluate(self, schedule):
        """ Machine selects jobs until all jobs have a machine """
        raise NotImplementedError


class Job(object):
    ids = count()

    def __init__(self):
        self.id = next(Job.ids)
        self.paths = Graph()
        self.candidate_machines = []
        self.schedule = []

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "{}-{}".format(self.__class__.__name__, self.id)

    def __hash__(self):
        return self.id

    def add_appointment(self, appointment):
        self.schedule.append(appointment)

    def is_scheduled(self):
        if len(self.paths.nodes()) == 0:
            raise ValueError("add process first.")

        if len(self.paths.nodes()) == 1:
            proc_id = self.paths.nodes()[0]
            if len(self.schedule) == 1:
                appointment = self.schedule[0]
                if appointment.process == proc_id:
                    return True

        path = [app.process for app in self.schedule]
        if path[0] not in self.paths.nodes(in_degree=0):  # starts
            return False
        if path[-1] not in self.paths.nodes(out_degree=0):  # ends
            return False
        if not self.paths.has_path(path):
            return False
        item_1 = self.schedule[0]
        for item_2 in self.schedule[1:]:
            if item_1.finish > item_2.start:
                raise AppointmentError(f"{item_2.start} start before {item_1.finish} ends")
            item_1 = item_2
        return True

    def add_process(self, steps):
        """
        :param: steps: a list of required processing steps.

        minimum: [1] interpreted as "one step" process.

        to provide multiple paths, for example [1,2,3] or [1,4,3]
        simply add processes twice, as the job creates a graph of
        options as:

        [1] ---> [2] ---> [3]
           +             /
            ---> [4] -->
        """
        assert isinstance(steps, list)
        if len(steps) == 1:
            self.paths.add_node(steps[0])
        for idx in range(len(steps) - 1):
            self.paths.add_edge(steps[idx], steps[idx + 1])

    def evaluate(self, schedule):
        """ Job selets machines until all process steps have a machine """
        raise NotImplementedError


def minimise_makespan(machines, jobs):
    """
    Minimises the makespan of the scheduling problem
    :param machines: list of machines
    :param jobs: list of jobs
    :return list of schedules.  # :return: tuple(list of machines, list of jobs)

    by returning the same output, as given as input,
    multiple iterations can be performed.
    """
    assert isinstance(machines, list)
    assert all(isinstance(i, Machine) for i in machines)
    assert isinstance(jobs, list)
    assert all(isinstance(i, Job) for i in jobs)

    makespan_min = float('inf')  # the default value for an unsolved scheduling problem.
    solutions = set()  # collection of graph-hashes for explored solutions.
    best_solution = None  # the solution with the smallest makespan.
    compatibility = Graph()  # bi-directional di-graph with machines and jobs based on processes.
    for m in machines:
        m_proc = set(m.processes.keys())  # processes that the machine supports.
        for j in jobs:
            j_procs = set(j.paths.nodes())  # processes that the job requires.
            if j_procs.intersection(m_proc):
                compatibility.add_edge(m, j)
                compatibility.add_edge(j, m)

    # check that all jobs can be processed.
    all_procs = set([m.processes.keys() for m in machines])
    all_jobs = set([j.paths.nodes() for j in jobs])
    if all_jobs.issuperset(all_procs):
        missing = all_jobs.difference(all_procs)
        raise ValueError("machines do not support processes: {}".format(missing))

    # machines offer services to jobs.
    #     candidates = compatibility.nodes(from_node=m)
    #     m.candidate_jobs()

    while True:
        # 1. let jobs book time with machines.
        # 2. let machines move jobs around to reduce idle time.
        # 3. evaluate current performance...
        ms = makespan(machines, jobs)  # evaluate the performance.
        if ms == float('inf'):  # means that there are still unassigned jobs.
            continue

        if ms < makespan_min:  # remember the solution if the performance is better than best previously known solution.
            makespan_min = ms
            best_solution = [m.schedule.copy() for m in machines]

        gh = graph_hash(solution(machines, jobs))
        if gh in solutions:  # the solution search is cyclic: abort with best known solution.
            return best_solution
        else:
            solutions.add(gh)


def makespan(machines, jobs):  # todo
    """ calculates the makespan of the schedule.
    :param: machines: list of machines
    :param: jobs: list of jobs
    :return: float: the makespan of the assignment.
    If all jobs are not assigned, float('inf') is returned.

    """
    # 1. make a set of all jobs and check if they're scheduled.
    if not all(job.is_scheduled() for job in jobs):
        return float('inf')
    # 2. find the max finish time of all machines
    return max((m.makespan_finish() for m in machines))


def solution(machines, jobs):
    """
    :param machines: list of machines
    :param jobs: list of jobs
    :return: graph of assignments
    """
    raise NotImplementedError


def gantt(machines, jobs):
    """ creates a gantt chart of the machines and their assigned jobs. """
    assert isinstance(machines, list), 'expected list of Machines'
    assert all(isinstance(i, Machine) for i in machines), 'expected list of Machines'
    assert isinstance(jobs, list), 'expected list of Jobs'
    assert all(isinstance(i, Job) for i in jobs), 'Expected list of Jobs'

    fig, ax = plt.subplots()

    # to debug see: https://matplotlib.org/examples/pylab_examples/broken_barh.html
    x_max = 0
    for m in machines:
        xs = []
        for s in m.schedule:
            xs.append((s.start, s.end - s.start))
            x_max = max(x_max, s.end)
        ys = (m.id * 10, 8)  # y offset, y height
        ax.broken_barh(xs, ys)

        for s in m.schedule:
            ax.annotate('J{}.{}'.format(s.job, s.proc), (s.start, m.id * 10),
                        fontsize=8,
                        horizontalalignment='right', verticalalignment='top')

    ax.set_ylim(0, len(machines) * 10 + 10)
    ax.set_xlim(0, x_max)
    ax.set_xlabel('time')
    ax.set_yticks([(m.id * 10) - 5 for m in machines])
    ax.set_yticklabels([m.id for m in machines])
    ax.grid(True)

    plt.show()





from graph import Graph
from itertools import count


class CVRPTW(object):
    """ solves the Capacitated Vehicle Routing Problem with Time Windows. """

    def __init__(self, graph):
        assert isinstance(graph, Graph)
        self.graph = graph
        self.am = self.graph.adjacency_matrix()
        self.vehicles = []
        self.loads = []

    def add(self, item):
        if isinstance(item, Load):
            self._new_load(item)
        elif isinstance(item, Vehicle):
            self._new_vehicle(item)
        else:
            raise NotImplementedError(f"no method for {type(item)}")

    def distance(self, a, b):
        return self.am[a][b]

    def _new_load(self, load):
        assert isinstance(load, Load)
        if load.from_loc not in self.graph.nodes():
            raise ValueError("start not in map")
        if load.to_loc not in self.graph.nodes():
            raise ValueError("end not in map")
        self.loads.append(load)

    def _new_vehicle(self, vehicle):
        assert isinstance(vehicle, Vehicle)
        if vehicle.start not in self.graph.nodes():
            raise ValueError("start not in map")
        self.vehicles.append(vehicle)
        vehicle.pd = self

    def solve(self):
        """ top level method for calling the solver"""
        # 1. create initial solution
        for vehicle in self.vehicles:
            vehicle.evaluate()
        # 2. swop until no further improvement
        pass

    def __str__(self):
        s = []
        sa = s.append
        for vehicle in self.vehicles:
            sa(str(vehicle))
            for tour in vehicle.tour:
                sa(str(tour))
        return "\n".join(s)


class Vehicle(object):
    id = count(1)

    def __init__(self, start, capacity, endurance, return_to_start=True):
        """
        :param start: Start location
        :param capacity: Vehicle capacity
        :param endurance: Time vehicle is available.
        """
        self.uid = next(Vehicle.id)
        self.start = start
        self.capacity = capacity
        self.used_capacity = 0
        self.endurance = endurance
        self.used_endurance = 0
        self._must_return_to_start = return_to_start
        load = Load(start, end=start, size=0, open=0, close=self.endurance, handling_time=0)
        self.tour = [load]
        self.pd = None  # place holder from ProblemDefinition (pd)

    def __str__(self):
        return f"Vehicle {self.uid}:"

    def tour_length(self):
        start = self.tour[0]
        t = 0
        for end in self.tour[1:]:
            load = end.handling_time
            unload = end.handling_time
            travel = self.pd.am[start.to_loc][end.to_loc]
            t += load + unload + travel
        return t

    def evaluate(self):
        if self.pd is None:
            raise AttributeError("forgot to add the vehicle? Use: cvrptw.add(vehicle)")
        assert isinstance(self.pd, CVRPTW)
        # 1. make a shortlist:
        ranking = [load for load in self.pd.loads if load.vehicle is None]

        while ranking:
            # 2. rank the shortlisted loads
            last_load = self.tour[-1]
            tour_length = self.tour_length()
            used_capacity = sum([l.size for l in self.tour])

            new_rank = []
            add = new_rank.append
            for load in ranking:
                assert isinstance(load, Load)
                if load.vehicle is not None:
                    continue
                if load.size + used_capacity > self.capacity:
                    continue
                leg = self.pd.am[last_load.to_loc][load.to_loc]
                if tour_length + leg > self.endurance:
                    continue
                if self._must_return_to_start:
                    home_leg = self.pd.am[load.to_loc][self.start]
                    if tour_length + leg + home_leg > self.endurance:
                        continue

                add((load.open, load.close, leg, load))

            if new_rank:
                new_rank.sort()
                ranking = [load for o, c, d, load in new_rank]
                load = ranking.pop(0)  # 3. pop top ranking load
                load.vehicle = self
                self.tour.append(load)
            else:
                break

        if self._must_return_to_start:
            self.create_home_trip()

    def create_home_trip(self):
        """ adds return trip to start after end of search. """
        self.tour.append(self.tour[0])


class Load(object):
    id = count(1)

    def __init__(self, start, end, size, open, close, handling_time):
        """
        :param start: start location
        :param end: end location
        :param size: Size (relative to vehicle capacity)
        :param open: Time window opening
        :param close: Time window closing
        :param time: Time to load/unload
        """
        self.uid = next(Load.id)
        self.from_loc = start
        self.to_loc = end
        self.size = size
        self.open = open
        self.close = close
        self.vehicle = None
        self.handling_time = handling_time

    def __str__(self):
        return f"Load {self.uid}: from {self.from_loc} to {self.to_loc}"

    def __lt__(self, other):
        return self.open < other.open

    def __le__(self, other):
        return self.open <= other.open





class BasicGraph(object):
    """
    Graph is the base graph that all methods use.

    For methods, please see the documentation on the
    individual functions, by importing them separately.

    """

    def __init__(self, from_dict=None, from_list=None):
        """
        :param from_dict: creates graph for dictionary {n1:{n2:d} ...
        :param links: creates graph from list of edges(n1,n2,d)
        """
        self._nodes = {}
        self._edges = {}
        self._max_edge_value = 0

        if from_dict is not None:
            self.from_dict(from_dict)
        elif from_list is not None:
            self.from_list(from_list)

    def __getitem__(self, item):
        raise ValueError("Use g.node(n1) or g.edge(n1,n2)")

    def __setitem__(self, key, value):
        raise ValueError("Use add_edge(node1, node2, value)")

    def __delitem__(self, key):
        raise ValueError("Use del_edge(node1, node2)")

    def __contains__(self, item):
        """
        :returns bool: True if node in Graph.
        """
        return item in self._nodes

    def __len__(self):
        raise ValueError("Use len(g.nodes()) or len(g.edges())")

    def __copy__(self):
        if not self.__class__.__name__ == BasicGraph.__name__:
            raise NotImplementedError("subclasses must implement this method.")
        g = BasicGraph()
        for n in self._nodes:
            g.add_node(n, obj=self._nodes[n])
        for s, e, d in self.edges():
            g.add_edge(s, e, d)
        return g

    def add_edge(self, node1, node2, value=1, bidirectional=False):
        """
        :param node1: hashable node
        :param node2: hashable node
        :param value: numeric value (int or float)
        :param bidirectional: boolean.
        """
        if isinstance(value, (dict, list, tuple)):
            raise ValueError("value cannot be {}".format(type(value)))
        if node1 not in self._nodes:
            self.add_node(node1)
        if node2 not in self._nodes:
            self.add_node(node2)

        if node1 not in self._edges:
            self._edges[node1] = {}
        if node2 not in self._edges:
            self._edges[node2] = {}
        self._edges[node1][node2] = value
        if value > self._max_edge_value:
            self._max_edge_value = value
        if bidirectional:
            self._edges[node2][node1] = value

    def edge(self, node1, node2, default=None):
        """Retrieves the edge (node1, node2)

        Alias for g[node1][node2]

        :param node1: node id
        :param node2: node id
        :param default: returned value if edge doesn't exist.
        :return: edge(node1,node2)
        """
        try:
            return self._edges[node1][node2]
        except KeyError:
            return default

    def del_edge(self, node1, node2):
        """
        removes edge from node1 to node2
        :param node1: node
        :param node2: node
        """
        del self._edges[node1][node2]

    def add_node(self, node_id, obj=None):
        """
        :param node_id: any hashable node.
        :param obj: any object that the node should refer to.

        PRO TIP: To retrieve the node obj use g.node(node_id)

        """
        self._nodes[node_id] = obj

    def node(self, node_id):
        """
        Retrieves the node object

        :param node_id: id of node in graph.
        :return: node object
        """
        return self._nodes.get(node_id, None)

    def del_node(self, node_id):
        """
        Deletes the node and all its connections.
        :param node_id: node_id
        :return: None
        """
        try:
            del self._nodes[node_id]
        except KeyError:
            pass
        try:
            del self._edges[node_id]
        except KeyError:
            pass
        in_links = [n1 for n1, n2, d in self.edges() if n2 == node_id]
        for inlink in in_links:
            del self._edges[inlink][node_id]
        return None

    def nodes(self,
              from_node=None, to_node=None,
              in_degree=None, out_degree=None):
        """
        :param from_node (optional) return nodes with edges from 'from_node'
        :param to_node (optional) returns nodes with edges into 'to_node'
        :param in_degree (optional) returns nodes with in_degree=N
        :param out_degree (optional) returns nodes with out_degree=N

        :return list of node ids.
        """
        inputs = sum([1 for i in (from_node, to_node, in_degree, out_degree) if i is not None])
        if inputs > 1:
            m = []
            a = (from_node, to_node, in_degree, out_degree)
            b = ("from_node", "to_node", "in_degree", "out_degree")
            for i in zip(a, b):
                if i is not None:
                    m.append("{}={}".format(b, a))
            raise ValueError("nodes({}) has too many inputs. Pick one.".format(m))

        if inputs == 0:
            return list(self._nodes.keys())

        if from_node is not None:
            # return [n2 for n1, n2, d in self.edges(from_node=from_node)]
            if self._edges.get(from_node, None) is not None:
                return [n2 for n2, v in self._edges[from_node].items()]
            return []

        if to_node is not None:
            return [n1 for n1, n2, d in self.edges() if n2 == to_node]

        if in_degree is not None:
            if not isinstance(in_degree, int) or in_degree < 0:
                raise ValueError("in_degree must be int >= 0")

            rev = {n: set() for n in self._nodes}
            for n1, n2, d in self.edges():
                rev[n2].add(n1)
            return [n for n, n_set in rev.items() if len(n_set) == in_degree]

        if out_degree is not None:
            if not isinstance(out_degree, int) or out_degree < 0:
                raise ValueError("out_degree must be int >= 0")

            rev = {n: set() for n in self._nodes}
            for n1, n2, d in self.edges():
                rev[n1].add(n2)
            return [n for n, n_set in rev.items() if len(n_set) == out_degree]

    def edges(self, path=None, from_node=None, to_node=None):
        """
        :param path (optional) list of nodes for which the edges are wanted.
        :param from_node (optional) for which outgoing edges are returned.
        :param to_node (optiona) for which incoming edges are returned.
        :return list of edges (n1, n2, value)
        """
        inputs = sum([1 for i in (from_node, to_node, path) if i is not None])
        if inputs > 1:
            m = []
            a = (path, from_node, to_node)
            b = ("path", "from_node", "to_node")
            for i in zip(a, b):
                if i is not None:
                    m.append("{}={}".format(b, a))
            raise ValueError("edges({}) has too many inputs. Pick one.".format(m))

        if path:
            if not isinstance(path, list):
                raise ValueError("expects a list")
            if len(path) < 2:
                raise ValueError("path of length 1 is not a path.")

            return [(path[ix], path[ix + 1], self._edges[path[ix]][path[ix + 1]])
                    for ix in range(len(path ) -1)]

        if from_node:
            if from_node in self._edges:
                return [(from_node, n2, self._edges[from_node][n2]) for n2 in self._edges[from_node]]
            else:
                return []

        if to_node:
            return [(n1, n2, self._edges[n1][n2])
                    for n1 in self._edges
                    for n2 in self._edges[n1]
                    if n2 == to_node]

        return [(n1, n2, self._edges[n1][n2]) for n1 in self._edges for n2 in self._edges[n1]]

    def from_dict(self, dictionary):
        """
        Updates the graph from dictionary
        :param dictionary:

        d = {1: {2: 10, 3: 5},
             2: {4: 1, 3: 2},
             3: {2: 3, 4: 9, 5: 2},
             4: {5: 4},
             5: {1: 7, 4: 6}}

        G = Graph(from_dict=d)

        :return: None
        """
        assert isinstance(dictionary, dict)
        for n1, e in dictionary.items():
            if not e:
                self.add_node(n1)
            else:
                for n2, v in e.items():
                    self.add_edge(n1, n2, v)

    def to_dict(self):
        """ creates a nested dictionary from the graph.
        :return dict d[n1][n2] = distance
        """
        d = {}
        for n1, n2, dist in self.edges():
            if n1 not in d:
                d[n1] = {}
            d[n1][n2] = dist

        for n in self.nodes():
            if n not in d:
                d[n] = {}
        return d

    def from_list(self, links):
        """
        updates the graph from a list of links.
        :param links: list

        links = [
            (1, 2, 18),
            (1, 3, 10),
            (2, 4, 7),
            (2, 5, 6),
            (3, 4, 2),
            (11,)      # node with no links.
        ]
        """
        assert isinstance(links, list)
        for item in links:
            assert isinstance(item, tuple)
            if len(item) == 3:
                self.add_edge(*item)
            else:
                self.add_node(item[0])

    def to_list(self):
        """ returns list of edges and nodes."""
        return self.edges() + [(i,) for i in self.nodes()]




from graph import BasicGraph


def adjacency_matrix(graph):
    """
    :param graph:
    :return: dictionary

    Converts directed graph to an adjacency matrix.
    Note: The distance from a node to itself is 0 and distance from a node to
    an unconnected node is defined to be infinite. This does not mean that there
    is no path from a node to another via other nodes.

    Example:
        g = Graph(from_dict=
            {1: {2: 3, 3: 8, 5: -4},
             2: {4: 1, 5: 7},
             3: {2: 4},
             4: {1: 2, 3: -5},
             5: {4: 6}})

        adjacency_matrix(g)
        {1: {1: 0, 2: 3, 3: 8, 4: inf, 5: -4},
         2: {1: inf, 2: 0, 3: inf, 4: 1, 5: 7},
         3: {1: inf, 2: 4, 3: 0, 4: inf, 5: inf},
         4: {1: 2, 2: inf, 3: -5, 4: 0, 5: inf},
         5: {1: inf, 2: inf, 3: inf, 4: 6, 5: 0}}
    """
    assert isinstance(graph, BasicGraph)
    return {v1: {v2: 0 if v1 == v2 else graph.edge(v1, v2, default=float('inf'))
                 for v2 in graph.nodes()}
            for v1 in graph.nodes()}


def all_pairs_shortest_paths(graph):
    """
    Find the cost of the shortest path between every pair of vertices in a
    weighted graph. Uses the Floyd-Warshall algorithm.

    Example:
        inf = float('inf')
        g = Graph(from_dict=(
            {0: {0: 0,   1: 1,   2: 4},
             1: {0: inf, 1: 0,   2: 2},
             2: {0: inf, 1: inf, 2: 0}})

        fw(g)
        {0: {0: 0,   1: 1,   2: 3},
        1: {0: inf, 1: 0,   2: 2},
        2: {0: inf, 1: inf, 2: 0}}

        h = {1: {2: 3, 3: 8, 5: -4},
             2: {4: 1, 5: 7},
             3: {2: 4},
             4: {1: 2, 3: -5},
             5: {4: 6}}

        fw(adj(h)) #
            {1: {1: 0, 2: 1, 3: -3, 4: 2, 5: -4},
             2: {1: 3, 2: 0, 3: -4, 4: 1, 5: -1},
             3: {1: 7, 2: 4, 3: 0, 4: 5, 5: 3},
             4: {1: 2, 2: -1, 3: -5, 4: 0, 5: -2},
             5: {1: 8, 2: 5, 3: 1, 4: 6, 5: 0}}
    """
    g = graph.adjacency_matrix()
    assert isinstance(g, dict)
    vertices = g.keys()

    for v2 in vertices:
        g = {v1: {v3: min(g[v1][v3], g[v1][v2] + g[v2][v3])
                  for v3 in vertices}
             for v1 in vertices}
    return g


from graph import BasicGraph
from graph.search import shortest_path


def maximum_flow(graph, start, end):
    """
    Returns the maximum flow graph
    :param graph: instance of Graph
    :param start: node
    :param end: node
    :return: flow, graph
    """
    assert isinstance(graph, BasicGraph)
    inflow = sum(d for s, e, d in graph.edges(from_node=start))
    outflow = sum(d for s, e, d in graph.edges(to_node=end))
    unassigned_flow = min(inflow, outflow)  # search in excess of this 'flow' is a waste of time.
    total_flow = 0
    # -----------------------------------------------------------------------
    # The algorithm
    # I reviewed a number of algorithms, such as Ford-fulkerson algorithm,
    # Edmonson-Karp and Dinic, but I didn't like them due to their naive usage
    # of BFS, which leads to a lot of node visits.
    #
    # I therefore choose to invert the capacities of the graph so that the
    # capacity any G[u][v] = c becomes 1/c in G_inverted.
    # This allows me to use the shortest path method to find the path with
    # most capacity in the first attempt, resulting in a significant reduction
    # of unassigned flow.
    #
    # By updating G_inverted, with the residual capacity, I can keep using the
    # shortest path, until the capacity is zero, whereby I remove the links
    # When the shortest path method returns 'No path' or when unassigned flow
    # is zero, I exit the algorithm.
    #
    # Even on small graphs, this method is very efficient, despite the overhead
    # of using shortest path. For very large graphs, this method outperforms
    # all other algorithms by orders of magnitude.
    # -----------------------------------------------------------------------

    edges = [(n1, n2, 1 / d) for n1, n2, d in graph.edges() if d > 0]
    inverted_graph = BasicGraph(from_list=edges)  # create G_inverted.
    capacity_graph = BasicGraph()  # Create structure to record capacity left.
    flow_graph = BasicGraph()  # Create structure to record flows.

    while unassigned_flow:
        # 1. find the best path
        d, path = shortest_path(inverted_graph, start, end)
        if d == float('inf'):  # then there is no path, and we must exit.
            return total_flow, flow_graph
        # else: use the path and lookup the actual flow from the capacity graph.

        path_flow = min([min(d, capacity_graph.edge(s, e, default=float('inf')))
                         for s, e, d in graph.edges(path=path)])

        # 2. update the unassigned flow.
        unassigned_flow -= path_flow
        total_flow += path_flow

        # 3. record the flows and update the inverted graph, so that it is
        #    ready for the next iteration.
        edges = graph.edges(path)
        for n1, n2, d in edges:

            # 3.a. recording:
            v = flow_graph.edge(n1, n1, default=None)
            if v is None:
                flow_graph.add_edge(n1, n2, path_flow)
                c = graph.edge(n1, n2) - path_flow
            # else:  <-- Note I doubt this path is ever reached.
            #     flow_graph.add_edge(n1, n2, value=v + path_flow)
            #     c = graph.edge(n1, n2) - (v + path_flow)
            capacity_graph.add_edge(n1, n2, c)

            # 3.b. updating:
            # if there is capacity left: update with new 1/capacity
            # else: remove node, as we can't do 1/zero.
            if c > 0:
                inverted_graph.add_edge(n1, n2, 1 / c)
            else:
                inverted_graph.del_edge(n1, n2)
    return total_flow, flow_graph




import hashlib

from graph import Graph


def graph_hash(graph):
    """ Generates the top hash of the graph.
    :param graph: instance of class Graph.
    :return: graph hash (int) and graph (Graph) with hash values
    """
    assert isinstance(graph, Graph)
    hash_func = hashlib.sha3_256()
    nodes = bytes("|".join(str(n) for n in sorted(graph.nodes())), 'utf-8')
    edges = bytes("|".join(str(e) for e in sorted(graph.edges())), 'utf-8')
    hash_func.update(nodes + edges)
    return int(hash_func.hexdigest(), 16)


def flow_graph_hash(graph):
    """
    Calculates the hash of a flow graph, where the properties of the nodes
    and the supplying nodes are included in the hash.

    Any upstream change in hash will thereby propagate downstream.
    """
    assert isinstance(graph, Graph)
    sources = graph.nodes(in_degree=0)

    original_hash = 'original hash'
    new_hash = 'new_hash'
    hash_graph = Graph()  # new graph with hashes.
    visited = set()

    while sources:
        source = sources[0]
        sources = sources[1:]

        suppliers = graph.nodes(to_node=source)

        hash_func = hashlib.sha3_256()
        hash_func.update(bytes(str(source), 'utf-8'))
        for supplier in suppliers:
            if graph.depth_first_search(start=source, end=supplier):
                continue  # it's a cycle.
            d = hash_graph.node(supplier)
            hash_func.update(bytes(d[new_hash], 'utf-8'))
        source_hash = hash_func.hexdigest()

        if source not in hash_graph:
            obj = {original_hash: source, new_hash: source_hash}
            hash_graph.add_node(source, obj=obj)
        else:
            n = hash_graph.node(source)
            n[new_hash] = source_hash

        receivers = graph.nodes(from_node=source)
        for receiver in receivers:
            if receiver in visited:
                continue
            visited.add(receiver)

            if receiver not in hash_graph:
                obj = {original_hash: receiver, new_hash: None}
                hash_graph.add_node(node_id=receiver, obj=obj)
            hash_graph.add_edge(source, receiver)
            if receiver not in sources:
                sources.append(receiver)

    for sink in graph.nodes(out_degree=0):
        n = hash_graph.node(sink)
        assert n[new_hash] is not None, n

    return hash_graph


def merkle_tree(data_blocks):
    """

A hash tree or Merkle tree is a tree in which every leaf node is labelled with
the hash of a data block, and every non-leaf node is labelled with the
cryptographic hash of the labels of its child nodes. Hash trees allow efficient
and secure verification of the contents of large data structures. Hash trees
are a generalization of hash lists and hash chains.

                    Top Hash
                 hash ( 0 + 1 )
                   ^      ^
                   |      |
           +------->      +---------+
           ^                        ^
           |                        |
           +                        +
        Hash 0                    Hash 1
  hash ( 0-0 + 0-1 )        hash ( 1-0 + 1-1 )
     ^           ^           ^           ^
     |           |           |           |
     +           +           +           +
  Hash 0-0    Hash 0-1    Hash 1-0     Hash 1-1
  hash(L1)    hash(L2)    hash(L3)     hash(L4)
     ^           ^           ^           ^
     |           |           |           |
+----------------------------------------------+
|    L1          L2          L3          L4    | Data blocks
+----------------------------------------------+

    """
    g = Graph()

    # initial hash:
    leaves = []
    for block in data_blocks:
        assert isinstance(block, bytes)
        hash_func = hashlib.sha3_256()
        hash_func.update(block)
        uuid = hash_func.hexdigest()
        leaves.append(uuid)
        g.add_node(node_id=uuid)

    # populate graph
    while leaves:
        if len(leaves) == 1:
            return g  # <--- point of return.

        c1, c2 = leaves[:2]
        leaves = leaves[2:]

        hash_func = hashlib.sha3_256()
        hash_func.update(bytes(c1, 'utf-8') + bytes(c2, 'utf-8'))
        uuid = hash_func.hexdigest()
        leaves.append(uuid)
        g.add_node(node_id=uuid)
        g.add_edge(c1, uuid)
        g.add_edge(c2, uuid)


from collections import defaultdict
from functools import lru_cache
from heapq import heappop, heappush
from itertools import combinations, chain

from graph.graphs import BasicGraph
from graph.transform import all_pairs_shortest_paths
from graph.topology import has_path


def shortest_path(graph, start, end):
    """
    :param graph: class Graph
    :param start: start node
    :param end: end node
    :return: distance, path (as list),
             returns float('inf'), [] if no path exists.
    """
    assert isinstance(graph, BasicGraph)

    g = defaultdict(list)
    for n1, n2, dist in graph.edges():
        g[n1].append((dist, n2))

    q, visited, mins = [(0, start, ())], set(), {start: 0}
    while q:
        (cost, v1, path) = heappop(q)
        if v1 not in visited:
            visited.add(v1)
            path = (v1, path)

            if v1 == end:  # exit criteria.
                L = []
                while path:
                    v, path = path[0], path[1]
                    L.append(v)
                L.reverse()
                return cost, L

            for dist, v2 in g.get(v1, ()):
                if v2 in visited:
                    continue
                prev = mins.get(v2, None)
                next_node = cost + dist
                if prev is None or next_node < prev:
                    mins[v2] = next_node
                    heappush(q, (next_node, v2, path))
    return float("inf"), []


def breadth_first_search(graph, start, end):
    """ Determines the path from start to end with fewest nodes.
    :param graph: class Graph
    :param start: start node
    :param end: end node
    :return: path
    """
    assert isinstance(graph, BasicGraph)

    g = defaultdict(list)
    for n1, n2, dist in graph.edges():
        g[n1].append((dist, n2))

    q, visited, mins = [(0, start, ())], set(), {start: 0}
    while q:
        (cost, v1, path) = heappop(q)
        if v1 not in visited:
            visited.add(v1)
            path = (v1, path)

            if v1 == end:  # exit criteria.
                L = []
                while path:
                    v, path = path[0], path[1]
                    L.append(v)
                L.reverse()
                return cost, L

            for dist, v2 in g.get(v1, ()):
                if v2 in visited:
                    continue
                prev = mins.get(v2, None)
                next_node = cost + 1
                if prev is None or next_node < prev:
                    mins[v2] = next_node
                    heappush(q, (next_node, v2, path))
    return float("inf"), []


def depth_first_search(graph, start, end):
    """
    Determines path from start to end using
    'depth first search' with backtracking.

    :param graph: class Graph
    :param start: start node
    :param end: end node
    :return: path as list of nodes.
    """
    assert start in graph, "start not in graph"
    assert end in graph, "end not in graph"
    q = [start]
    path = []
    visited = set()
    while q:
        n1 = q.pop()
        visited.add(n1)
        path.append(n1)
        if n1 == end:
            return path  # <-- exit if end is found.
        for n2 in graph.nodes(from_node=n1):
            if n2 in visited:
                continue
            q.append(n2)
            break
        else:
            path.remove(n1)
            while not q and path:
                for n2 in graph.nodes(from_node=path[-1]):
                    if n2 in visited:
                        continue
                    q.append(n2)
                    break
                else:
                    path = path[:-1]
    return None  # <-- exit if not path was found.


def tsp(graph):
    """
    Attempts to solve the traveling salesmans problem TSP for the graph.

    Runtime approximation: seconds = 10**(-5) * (points)**2.31
    Solution quality: Range 98.1% - 100% optimal.

    :param graph: instance of class Graph
    :return: tour_length, path
    """

    def shortest_links_first(graph):
        """ returns a list of (distance, node1, node2) with shortest on top."""
        c = combinations(graph.nodes(), 2)
        distances = [(graph.edge(a, b), a, b) for a, b in c]
        if None in [a for a, b, c in distances]:
            raise ValueError("TSP requires a fully connected graph.")
        distances.sort()
        return distances

    def join_endpoints(endpoints, a, b):
        """ Join segments [...,a] + [b,...] into one segment. Maintain `endpoints`.
        :param endpoints:
        :param a: node
        :param b: node
        :return:
        """
        a_seg, b_seg = endpoints[a], endpoints[b]
        if a_seg[-1] is not a:
            a_seg.reverse()
        if b_seg[0] is not b:
            b_seg.reverse()
        a_seg += b_seg
        del endpoints[a]
        del endpoints[b]
        endpoints[a_seg[0]] = endpoints[a_seg[-1]] = a_seg
        return a_seg

    def tsp_tour_length(graph, tour):
        """ The TSP tour length WITH return to the starting point."""
        return sum(graph.edge(tour[i - 1], tour[i]) for i in range(len(tour)))

    def improve_tour(graph, tour):
        assert tour, "no tour to improve?"
        while True:
            improvements = {reverse_segment_if_improvement(graph, tour, i, j)
                            for (i, j) in sub_segments(len(tour))}
            if improvements == {None} or len(improvements) == 0:
                return tour

    @lru_cache()
    def sub_segments(n):
        """ Return (i, j) pairs denoting tour[i:j] sub_segments of a tour of length N."""
        return [(i, i + length) for length in reversed(range(2, n))
                for i in reversed(range(n - length + 1))]

    def reverse_segment_if_improvement(graph, tour, i, j):
        """If reversing tour[i:j] would make the tour shorter, then do it."""
        # Given tour [...A,B...C,D...], consider reversing B...C to get [...A,C...B,D...]
        a, b, c, d = tour[i - 1], tour[i], tour[j - 1], tour[j % len(tour)]
        # are old links (ab + cd) longer than new ones (ac + bd)? if so, reverse segment.
        if graph.edge(a, b) + graph.edge(c, d) > graph.edge(a, c) + graph.edge(b, d):
            tour[i:j] = reversed(tour[i:j])
            return True

    # The core TSP solver:
    assert isinstance(graph, BasicGraph), type(graph)

    # 1. create a path using greedy algorithm (picks nearest peer)
    new_segment = []
    endpoints = {n: [n] for n in graph.nodes()}
    L = shortest_links_first(graph)
    for _, a, b in L:
        if a in endpoints and b in endpoints and endpoints[a] != endpoints[b]:
            new_segment = join_endpoints(endpoints, a, b)
            if len(new_segment) == len(graph.nodes()):
                break  # return new_segment
    assert len(new_segment) == len(graph.nodes()), "there's an unconnected component."
    first_tour = new_segment[:]
    first_path_length = tsp_tour_length(graph, first_tour)

    # 2. run improvement on the created path.
    improved_tour = improve_tour(graph, new_segment)
    assert set(graph.nodes()) == set(improved_tour)

    second_path_length = tsp_tour_length(graph, improved_tour)

    assert first_path_length >= second_path_length, "first path was better than improved tour?! {} {}".format(
        first_path_length, second_path_length
    )

    return second_path_length, improved_tour


def shortest_tree_all_pairs(graph):
    """
       'minimize the longest distance between any pair'

    Note: This algorithm is not shortest path as it jumps
    to a new branch when it has exhausted a branch in the tree.
    :return: path
    """
    assert isinstance(graph, BasicGraph)
    g = all_pairs_shortest_paths(graph)
    assert isinstance(g, dict)

    distance = float('inf')
    best_starting_point = -1
    # create shortest path gantt diagram.
    for start_node in g.keys():
        if start_node in g:
            dist = sum(v for k, v in g[start_node].items())
            if dist < distance:
                best_starting_point = start_node
            # else: skip the node as it's isolated.
    g2 = g[best_starting_point]  # {1: 0, 2: 1, 3: 2, 4: 3}

    inv_g2 = {}
    for k, v in g2.items():
        if v not in inv_g2:
            inv_g2[v] = set()
        inv_g2[v].add(k)

    all_nodes = set(g.keys())
    del g
    path = []
    while all_nodes and inv_g2.keys():
        v_nearest = min(inv_g2.keys())
        for v in inv_g2[v_nearest]:
            all_nodes.remove(v)
            path.append(v)
        del inv_g2[v_nearest]
    return path


def all_paths(graph, start, end):
    """ Returns all paths from start to end.
    :param graph: instance of Graph
    :param start: node
    :param end: node
    :return: list of paths
    """
    assert isinstance(graph, BasicGraph)
    options = set(graph.nodes()) - {start, end}

    # powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    s = list(options)
    L = []
    for combination in chain.from_iterable(combinations(s, r) for r in range(len(s) + 1)):
        path = [start] + list(combination) + [end]
        if has_path(graph, path):
            L.append(path)
    return L


def distance(graph, path):
    """ Calculates the distance for the path in graph
    :param graph: class Graph
    :param path: list of nodes
    :return: distance
    """
    assert isinstance(graph, BasicGraph)
    assert isinstance(path, list)
    cache = BasicGraph()
    path_length = 0
    for idx in range(len(path) - 1):
        n1, n2 = path[idx], path[idx + 1]
        d = cache.edge(n1, n2, default=None)
        if d is None:
            d, _ = shortest_path(graph, n1, n2)
            if d == float('inf'):
                return float('inf')
            cache.add_edge(n1, n2, value=d)
        path_length += d
    return path_length


def degree_of_separation(graph, n1, n2):
    """ Calculates the degree of separation between 2 nodes."""
    assert isinstance(graph, BasicGraph)
    assert n1 in graph.nodes()
    d, p = breadth_first_search(graph, n1, n2)
    return d


__all__ = ['Graph']
from itertools import count

from graph.graphs import BasicGraph
from graph.topology import (
    subgraph,
    is_subgraph,
    is_partite,
    same,
    has_path,
    has_cycles,
    components,
    phase_lines,
    sources
)
from graph.flow_problem import maximum_flow
from graph.search import (
    shortest_path,
    breadth_first_search,
    depth_first_search,
    tsp,
    shortest_tree_all_pairs,
    all_paths,
    distance
)
from graph.transform import adjacency_matrix, all_pairs_shortest_paths


class Graph(BasicGraph):
    """
    Graph is the base graph that all methods use.

    For methods, please see the documentation on the
    individual functions, by importing them separately.

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __copy__(self):
        g = Graph()
        for n in self._nodes:
            g.add_node(n, obj=self._nodes[n])
        for s, e, d in self.edges():
            g.add_edge(s, e, d)
        return g

    def shortest_path(self, start, end):
        """
        :param start: start node
        :param end: end node
        :return: distance, path as list
        """
        return shortest_path(graph=self, start=start, end=end)

    def breadth_first_search(self, start, end):
        """ Determines the path with fewest nodes.
        :param start: start node
        :param end: end nodes
        :return: nodes, path as list
        """
        return breadth_first_search(graph=self, start=start, end=end)

    def depth_first_search(self, start, end):
        """
        Finds a path from start to end using DFS.
        :param start: start node
        :param end: end node
        :return: path
        """
        return depth_first_search(graph=self, start=start, end=end)

    def distance_from_path(self, path):
        """
        :param path: list of nodes
        :return: distance along the path.
        """
        return distance(graph=self, path=path)

    def maximum_flow(self, start, end):
        """ Determines the maximum flow of the graph between
        start and end.
        :param start: node (source)
        :param end: node (sink)
        :return: flow, graph of flow.
        """
        return maximum_flow(self, start, end)

    def solve_tsp(self):
        """ solves the traveling salesman problem for the graph
        (finds the shortest path through all nodes)
        :return: tour length (path+retrun to starting point),
                 path travelled.
        """
        return tsp(self)

    def subgraph_from_nodes(self, nodes):
        """
        constructs a copy of the graph containing only the
        listed nodes (and their links)
        :param nodes: list of nodes
        :return: class Graph
        """
        return subgraph(graph=self, nodes=nodes)

    def is_subgraph(self, other):
        """ Checks if self is a subgraph in other.
        :param other: instance of Graph
        :return: boolean
        """
        return is_subgraph(self, other)

    def is_partite(self, n=2):
        """ Checks if self is n-partite
        :param n: int the number of partitions.
        :return: tuple: boolean, partitions as dict
                        (or None if graph isn't n-partite)
        """
        return is_partite(self, n)

    def has_cycles(self):
        """ Checks if the graph has a cycle
        :return: bool
        """
        return has_cycles(graph=self)

    def components(self):
        """ Determines the number of components
        :return: list of sets of nodes. Each set is a component.
        """
        return components(graph=self)

    def phase_lines(self):
        """ Determines the phase lines (cuts) of the graph
        :returns: dictionary with phase: nodes in phase
        """
        return phase_lines(self)

    def sources(self, n):
        """ Determines the DAG sources of node n """
        return sources(graph=self, n=n)

    @staticmethod
    def same_path(p1, p2):
        """ compares two paths to determine if they're the same, despite
        being in different order.
        :param p1: list of nodes
        :param p2: list of nodes
        :return: boolean
        """
        return same(p1, p2)

    def adjacency_matrix(self):
        """
        Converts directed graph to an adjacency matrix.
        Note: The distance from a node to itself is 0 and distance from a node to
        an unconnected node is defined to be infinite. This does not mean that there
        is no path from a node to another via other nodes.
        :return: dict
        """
        return adjacency_matrix(graph=self)

    def all_pairs_shortest_paths(self):
        """
        Find the cost of the shortest path between every pair of vertices in a
        weighted graph. Uses the Floyd-Warshall algorithm.
        :return: dict {node 1: {node 2: distance}, ...}
        """
        return all_pairs_shortest_paths(graph=self)

    def shortest_tree_all_pairs(self):
        """
        :return:
        """
        return shortest_tree_all_pairs(graph=self)

    def has_path(self, path):
        """
        :param path: list of nodes
        :return: boolean, if the path is in G.
        """
        return has_path(graph=self, path=path)

    def all_paths(self, start, end):
        """
        finds all paths from start to end
        :param start: node
        :param end: node
        :return: list of paths
        """
        return all_paths(graph=self, start=start, end=end)


class Graph2D(BasicGraph):
    nid = count()
    """ a graph where all (x,y)-positions are unique. """
    def __init__(self, from_xy_dict=None, from_xy_list=None):
        super().__init__()
        self.xys = {}
        if from_xy_dict is not None:
            self.from_xy_dict(dictionary=from_xy_dict)
        if from_xy_list is not None:
            self.from_xy_list(links=from_xy_list)

    def __copy__(self):
        g = Graph2D(from_xy_dict=self.to_xy_dict())
        return g

    def add_xy_edge(self, x1, y1, x2, y2, value, bidirectional=False):
        nid_1 = self.xys.get((x1, y1), None)
        if nid_1 is None:
            nid_1 = self.add_xy_node(x1, y1)
        nid_2 = self.xys.get((x2, y2), None)
        if nid_2 is None:
            nid_2 = self.add_xy_node(x2, y2)

        self.add_edge(nid_1, nid_2, value=value, bidirectional=bidirectional)

    def add_xy_node(self, x, y, node_id=None):
        """ Adds node from xy coordinates """
        obj = (x, y)
        if obj in self.xys:
            raise ValueError("({},{}) already exists".format(x, y))
        if node_id is None:
            node_id = next(Graph2D.nid)
        self.add_node(node_id, obj)
        self.xys[obj] = node_id
        return node_id

    def node_id(self, x, y):
        return self.xys.get((x, y), None)

    def from_xy_dict(self, dictionary):
        """
        Updates the graph from dictionary
        :param dictionary

        d = {(1, 2): {(1, 7): 5,
                      (6, 2): 5},
             (1, 7): {(4, 4): 4,
                      (6, 2): 9},
             (6, 2): {(1, 7): 9,
                      (4, 4): 5,
                      (8, 5): 5.6},
             (4, 4): {(8, 5): 4},
             (8, 5): {(1, 2): 7,
                      (4, 4): 6}}
        G = Graph2D(from_dict=d)
        """
        def check(xy):
            assert isinstance(xy, tuple)
            assert len(xy) == 2
            assert all((isinstance(a, (float, int)) for a in xy))

        for k, v in dictionary.items():
            check(k)
            if k not in self.xys:
                x1, y1 = k
                nid_1 = self.add_xy_node(x1, y1)
            else:
                nid_1 = self.xys[k]
            for k2, v2 in v.items():
                check(k2)
                assert isinstance(v2, (int, float))
                if k2 not in self.xys:
                    x2, y2 = k2
                    nid_2 = self.add_xy_node(x2, y2)
                else:
                    nid_2 = self.xys[k2]
                self.add_edge(nid_1, nid_2, value=v2)

    def to_xy_dict(self):
        """ creates a nested dictionary from the graph.
        :return dict d[xy_1][xy_2] = distance
        """
        d = {}
        for n1, n2, dist in self.edges():
            xy1 = self._nodes[n1]
            xy2 = self._nodes[n2]
            if xy1 not in d:
                d[xy1] = {}
            d[xy1][xy2] = dist

        for n in self.nodes():
            xy = self._nodes[n]
            if xy not in d:
                d[xy] = {}
        return d

    def from_xy_list(self, links):
        """
        updates the graph from a list of xy coordinate pairs
        :param links:

        links = [
            ((1,2),(3,4),5),
            ((1,2),(4,3),5),
            ((2,4),(7,1),7),
            ((3,3),(7,1),7),
            ((4,4),)
        ]
        """
        assert isinstance(links, list)
        for item in links:
            assert isinstance(item, tuple)
            if len(item) == 3:
                self.add_xy_edge(*item)
            else:
                self.add_xy_node(*item[0])

    def to_xy_list(self):
        """ returns list of edges and nodes."""
        def xy(nid):
            return self.node(nid)

        return [(xy(s), xy(e), d) for s, e, d in self.edges()] + [(xy(i),) for i in self.nodes()]

    @staticmethod
    def xy_distance(x1, y1, x2, y2):
        """ returns the distance between to xy coordinates"""
        return ((x1 - x2) * (x1 - x2) + (y1 - y2) * (y1 - y2)) ** (1 / 2)

    def n_nearest_neighbours(self, x, y, n=1):
        """ returns the node id of the `n` nearest neighbours. """
        dist = self.xy_distance
        d = [(dist(x, y, x1, y1), n) for (x1, y1), n in self.xys.items()]
        d.sort()
        if d:
            return [b for a, b in d][:n]
        return None

    def plot(self, nodes=True, links=True):
        """ plots nodes and links using matplotlib3"""
        raise NotImplementedError  # TODO see test_graph\test_random_graph_3 for plotting.


class Graph3D(BasicGraph):
    def __init__(self):
        super().__init__()
        # TODO